# Copyright (c) 2020 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2008 rule R502.

attr-spec is access-spec
             or ALLOCATABLE
             or ASYNCHRONOUS
             or CODIMENSION lbracket coarray-spec rbracket
             or CONTIGUOUS
             or DIMENSION ( component-array-spec )
             or EXTERNAL
             or INTENT ( intent-spec )
             or INTRINSIC
             or language-binding-spec
             or OPTIONAL
             or PARAMETER
             or POINTER
             or PROTECTED
             or SAVE
             or TARGET
             or VALUE
             or VOLATILE
"""

import pytest
from fparser.two.Fortran2008 import (
    Attr_Spec,
    Codimension_Attr_Spec,
    Deferred_Coshape_Spec,
    Coshape_Spec,
)
from fparser.two import Fortran2003
from fparser.api import get_reader


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize(
    "mixed_case_ref",
    [
        "ALlOCaTABLE",
        "ASYNChroNOUS",
        "CONTIGUOUs",
        "eXTERNAL",
        "INtRINsIC",
        "OPTIOnAL",
        "PARAmeTER",
        "PoInTeR",
        "pROTECTEd",
        "SAvE",
        "TargeT",
        "VALUE",
        "VOLATILE",
    ],
)
def test_attr_spec(mixed_case_ref):
    """Test the parsing of common attributes provided as mixed case, upper
    case and lower case strings."""
    for manip in [str, str.upper, str.lower]:
        obj = Attr_Spec(manip(mixed_case_ref))
        assert isinstance(obj, Attr_Spec), repr(obj)
        assert str(obj) == mixed_case_ref.upper()


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize(
    "attr", ["", "CONTIGUOS", " SAVE", "TARGET ", "VA LUE", "POINNTER", " DIMENSION(:)"]
)
def test_invalid_attr_spec(attr):
    """Test that invalid or misspelled attributes raise exceptions."""
    with pytest.raises(Fortran2003.NoMatchError):
        _ = Attr_Spec(attr)


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize(
    "attr, ref",
    [
        ("DIMENSION   (:)", "DIMENSION(:)"),
        ("dimenSION(  : )", "DIMENSION(:)"),
        ("DIMENSION (5)", "DIMENSION(5)"),
        ("DIMENSION (2:)", "DIMENSION(2 :)"),
        ("dimension (1:5, 2, 3:4)", "DIMENSION(1 : 5, 2, 3 : 4)"),
    ],
)
def test_attr_spec_dimension(attr, ref):
    """Test the parsing of dimension attribute in attr_spec."""
    obj = Attr_Spec(attr)
    assert isinstance(obj, Fortran2003.Dimension_Attr_Spec), repr(obj)
    assert str(obj) == ref


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize(
    "attr, ref",
    [
        ("codimension[*]", "CODIMENSION [*]"),
        ("CODIMENSION   [:]", "CODIMENSION [:]"),
        ("COdimenSION[  : ]", "CODIMENSION [:]"),
        ("coDIMENSIon [ *]", "CODIMENSION [*]"),
        ("CoDimension [1:*]", "CODIMENSION [1 : *]"),
        (" CODIMENSION [1:5, *]   ", "CODIMENSION [1 : 5, *]"),
        ("codimension [1:5, 2, 3:*]", "CODIMENSION [1 : 5, 2, 3 : *]"),
    ],
)
def test_attr_spec_codimension(attr, ref):
    """Test the parsing of codimension attribute in attr_spec."""
    obj = Attr_Spec(attr)
    assert isinstance(obj, Codimension_Attr_Spec), repr(obj)
    assert str(obj) == ref


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize(
    "attr",
    [
        "codimension[1:5]",
        "codimension[3:]",
        "codimension[:5]",
        "codimension[,,]",
        "codimension[1, 5, 3 *]",
        "co dimension[*]",
        "coodimension[*]",
        "codimension[]",
        "codimensions[*]",
        "codimension",
        "codimension[[*]",
        "codimension[[:]]",
        "codimension[][]",
    ],
)
def test_invalid_attr_spec_codimension(attr):
    """Test that invalid codimension attribute in attr_spec raise exception."""
    with pytest.raises(Fortran2003.NoMatchError):
        _ = Attr_Spec(attr)


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize(
    "attr, ref",
    [
        ("codimension[*]", "CODIMENSION [*]"),
        ("CODIMENSION   [:]", "CODIMENSION [:]"),
        ("COdimenSION[  : ]", "CODIMENSION [:]"),
        ("coDIMENSIon [ *]", "CODIMENSION [*]"),
        (" CODIMENSION [1:5, *]   ", "CODIMENSION [1 : 5, *]"),
        ("codimension [1:5, 2, 3:*]", "CODIMENSION [1 : 5, 2, 3 : *]"),
    ],
)
def test_codimension_attr_spec(attr, ref):
    """Test the parsing of codimension attributes."""
    obj = Codimension_Attr_Spec(attr)
    assert isinstance(obj, Codimension_Attr_Spec), repr(obj)
    assert str(obj) == ref


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize(
    "attr",
    [
        "codimension[1:5]",
        "codimension[3:]",
        "codimension[:5]",
        "codimension[,,]",
        "codimension[1, 5, 3 *]",
        "co dimension[*]",
        "coodimension[*]",
        "codimension[]",
        "codimensions[*]",
        "codimension",
        "codimension[[*]",
        "codimension[[:]]",
        "codimension[][]",
    ],
)
def test_invalid_codimension_attr_spec(attr):
    """Test that invalid codimension attributes raise exception."""
    with pytest.raises(Fortran2003.NoMatchError):
        _ = Codimension_Attr_Spec(attr)


def test_attr_spec_list_parser(f2008_parser):
    """Test that Attr_Spec_List is generated correctly and used by a parser."""
    code = """
SUBROUTINE FOO(i)
  INTEGER, CONTIGUOUS, INTENT(IN) :: i(10)
  INTEGER :: j
END SUBROUTINE
    """.strip()
    reader = get_reader(code)
    result = f2008_parser(reader)
    assert str(result) == code



# Copyright (c) 2020 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2008 rule R511

explicit-coshape-spec is [ coshape-spec-list , ] [ lower-cobound : ] *

"""

import pytest
from fparser.two.Fortran2008 import (
    Explicit_Coshape_Spec,
    Coshape_Spec,
    Coshape_Spec_List,
)
from fparser.two import Fortran2003


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize(
    "attr, types",
    [
        ("*", (type(None), type(None))),
        ("5 :*", (type(None), Fortran2003.Int_Literal_Constant)),
        ("1:  3  ,  a: *", (Coshape_Spec_List, Fortran2003.Name)),
        (" 1 ,  2,3:*", (Coshape_Spec_List, Fortran2003.Int_Literal_Constant)),
        ("1:2,3,*", (Coshape_Spec_List, type(None))),
    ],
)
def test_explicit_coshape_spec(attr, types):
    """Test that explicit_coshape_spec are parsed correctly."""
    obj = Explicit_Coshape_Spec(attr)
    assert isinstance(obj, Explicit_Coshape_Spec), repr(obj)
    assert isinstance(obj.items[0], types[0])
    assert isinstance(obj.items[1], types[1])
    ref = attr.replace(" ", "").replace(":", " : ").replace(",", ", ")
    assert str(obj) == ref


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize("attr", ["1:3", ":", ":b", "a:", ":*", "", "1:,*", "1,,*"])
def test_invalid_explicit_coshape_spec(attr):
    """Test that invalid explicit_coshape_spec raise exception."""
    with pytest.raises(Fortran2003.NoMatchError):
        _ = Explicit_Coshape_Spec(attr)


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize("attr", ["a", "10", "1:3", " 5 : 123 ", "2:   b   "])
def test_coshape_spec(attr):
    """Test that coshape_spec are parsed correctly."""
    obj = Coshape_Spec(attr)
    assert isinstance(obj, Coshape_Spec), repr(obj)
    assert str(obj) == attr.replace(" ", "").replace(":", " : ")


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize("attr", [":", "a:", ":b", "*", "", "1::3"])
def test_invalid_coshape_spec(attr):
    """Test that invalid coshape_spec raise exception."""
    with pytest.raises(Fortran2003.NoMatchError):
        _ = Coshape_Spec(attr)



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023-2024, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Fortran 2008 module. Contains classes which extend the Fortran
2003 standard to implement the Fortran 2008 standard.

"""

import inspect
import sys

from fparser.two.Fortran2003 import Base, SequenceBase
from fparser.two.Fortran2008.program_unit_r202 import Program_Unit
from fparser.two.Fortran2008.executable_construct_r213 import Executable_Construct
from fparser.two.Fortran2008.executable_construct_c201 import Executable_Construct_C201
from fparser.two.Fortran2008.action_stmt_r214 import Action_Stmt
from fparser.two.Fortran2008.action_stmt_c201 import Action_Stmt_C201
from fparser.two.Fortran2008.action_stmt_c816 import Action_Stmt_C816
from fparser.two.Fortran2008.action_stmt_c828 import Action_Stmt_C828
from fparser.two.Fortran2008.data_component_def_stmt_r436 import Data_Component_Def_Stmt
from fparser.two.Fortran2008.component_attr_spec_r437 import Component_Attr_Spec
from fparser.two.Fortran2008.type_declaration_stmt_r501 import Type_Declaration_Stmt
from fparser.two.Fortran2008.codimension_attr_spec_r502d import Codimension_Attr_Spec
from fparser.two.Fortran2008.coarray_bracket_spec_r502d0 import Coarray_Bracket_Spec
from fparser.two.Fortran2008.attr_spec_r502 import Attr_Spec
from fparser.two.Fortran2008.coarray_spec_r509 import Coarray_Spec
from fparser.two.Fortran2008.deferred_coshape_spec_r510 import Deferred_Coshape_Spec
from fparser.two.Fortran2008.explicit_coshape_spec_r511 import Explicit_Coshape_Spec
from fparser.two.Fortran2008.coshape_spec_r511a import Coshape_Spec
from fparser.two.Fortran2008.lower_cobound_r512 import Lower_Cobound
from fparser.two.Fortran2008.upper_cobound_r513 import Upper_Cobound
from fparser.two.Fortran2008.do_term_action_stmt_r826 import Do_Term_Action_Stmt
from fparser.two.Fortran2008.alloc_opt_r627 import Alloc_Opt
from fparser.two.Fortran2008.allocate_stmt_r626 import Allocate_Stmt
from fparser.two.Fortran2008.loop_control_r818 import Loop_Control
from fparser.two.Fortran2008.if_stmt_r837 import If_Stmt
from fparser.two.Fortran2008.error_stop_stmt_r856 import Error_Stop_Stmt
from fparser.two.Fortran2008.format_item_r1003 import Format_Item
from fparser.two.Fortran2008.stop_code_r857 import Stop_Code
from fparser.two.Fortran2008.specification_part_c1112 import Specification_Part_C1112
from fparser.two.Fortran2008.implicit_part_c1112 import Implicit_Part_C1112
from fparser.two.Fortran2008.implicit_part_stmt_c1112 import Implicit_Part_Stmt_C1112
from fparser.two.Fortran2008.declaration_construct_c1112 import (
    Declaration_Construct_C1112,
)
from fparser.two.Fortran2008.submodule_r1116 import Submodule
from fparser.two.Fortran2008.submodule_stmt_r1117 import Submodule_Stmt
from fparser.two.Fortran2008.end_submodule_stmt_r1119 import End_Submodule_Stmt
from fparser.two.Fortran2008.parent_identifier_r1118 import Parent_Identifier
from fparser.two.Fortran2008.open_stmt_r904 import Open_Stmt
from fparser.two.Fortran2008.connect_spec_r905 import Connect_Spec
from fparser.two.Fortran2008.block_construct_r807 import Block_Construct
from fparser.two.Fortran2008.block_stmt_r808 import Block_Stmt
from fparser.two.Fortran2008.end_block_stmt_r809 import End_Block_Stmt
from fparser.two.Fortran2008.critical_construct_r810 import Critical_Construct
from fparser.two.Fortran2008.critical_stmt_r811 import Critical_Stmt
from fparser.two.Fortran2008.end_critical_stmt_r812 import End_Critical_Stmt
from fparser.two.Fortran2008.procedure_stmt_r1206 import Procedure_Stmt

from fparser.two.Fortran2008.action_term_do_construct_r824 import (
    Action_Term_Do_Construct,
)
from fparser.two.Fortran2008.block_label_do_construct_r814_1 import (
    Block_Label_Do_Construct,
)
from fparser.two.Fortran2008.block_nonlabel_do_construct_r814_2 import (
    Block_Nonlabel_Do_Construct,
)
from fparser.two.Fortran2008.label_do_stmt_r816 import Label_Do_Stmt
from fparser.two.Fortran2008.nonlabel_do_stmt_r817 import Nonlabel_Do_Stmt

# pylint: disable=eval-used
# pylint: disable=exec-used

#
# GENERATE Scalar_, _List, _Name CLASSES
#
ClassType = type(Base)
_names = dir()
for clsname in _names:
    NEW_CLS = eval(clsname)
    if not (
        isinstance(NEW_CLS, ClassType)
        and issubclass(NEW_CLS, Base)
        and not NEW_CLS.__name__.endswith("Base")
    ):
        continue

    names = getattr(NEW_CLS, "subclass_names", []) + getattr(NEW_CLS, "use_names", [])
    for n in names:
        if n in _names:
            continue
        if n.endswith("_List"):
            _names.append(n)
            n = n[:-5]
            # Generate 'list' class
            exec(f"""\
class {n}_List(SequenceBase):
    subclass_names = [\'{n}\']
    use_names = []
    @staticmethod
    def match(string): return SequenceBase.match(r\',\', {n}, string)
""")
        elif n.endswith("_Name"):
            _names.append(n)
            n = n[:-5]
            exec(f"""\
class {n}_Name(Base):
    subclass_names = [\'Name\']
""")
        elif n.startswith("Scalar_"):
            _names.append(n)
            n = n[7:]
            exec(f"""\
class Scalar_{n}(Base):
    subclass_names = [\'{n}\']
""")
# Make sure NEW_CLS does not reference a class so is not accidentally
# picked up in __all__.
NEW_CLS = None


# Determine the generated classes in this module and list these in
# __all__ to support automatic documentation generation with AutoDoc.

classes = inspect.getmembers(
    sys.modules[__name__],
    lambda member: inspect.isclass(member) and member.__module__ == __name__,
)
__all__ = [name[0] for name in classes if name[0]]



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Science and Technology Facilities Council.
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Tests the parser against "do" block syntax.

This is a complicated piece of syntax with many axes of movement. A
comprehensive testing was felt to be too onerous so this stripped down
version checks only a subset. Hopefully it is representative.
"""

import pytest

from fparser.common.utils import AnalyzeError
from fparser.common.sourceinfo import FortranFormat
from fparser.one.parsefortran import FortranParser
from fparser.common.readfortran import FortranStringReader


def get_end_do(name):
    """A small helper function to return either "END DO" (without space) if
    name is empty, or "END DO "+name. This simplifies the tests now that
    tofortran does not return an "END DO" with one space in case of an
    unnamed end statement.

    :param str name: Either None if it is an unnamed statement, or \
        the label to use in the end statement.

    :returns: either "END DO" (without space) if name is empty, or \
        "END DO "+name.
    :rtype: str

    """

    if name:
        return "END DO {0}".format(name)
    return "END DO"


@pytest.mark.parametrize("name", [None, "loop_name"])
@pytest.mark.parametrize("label", [None, "123"])
@pytest.mark.parametrize("control_comma", [False, True])
@pytest.mark.parametrize(
    "terminal_expression", ["1", "10", "x+y", "size(array)", "size(this%array)"]
)
@pytest.mark.parametrize("end_name", [None, "loop_name", "wrong_name"])
@pytest.mark.parametrize("end_label", [None, "123", "456"])
def test_do(name, label, control_comma, terminal_expression, end_name, end_label):
    # pylint: disable=redefined-outer-name, too-many-arguments, too-many-locals
    """
    Checks that the "do" loop parser understands the "for-next" variant of
    the syntax. This is defined in BS ISO/IEC 1539-1:2010 with R814-R822.

    TODO: Only the terminal expression is tested. This is a short-cut and
          relies on expression handling being applied identically across
          all expressions. This was true at the time of writing the test.
    """
    name_snippet = name + ": " if name else None
    label_snippet = label + " " if label else None
    comma_snippet = ", " if control_comma else None
    # TODO: Although the Fortran standard allows for "continue" to be used in
    # place of "end do" fparser does not support it.
    end_snippet = "continue" if end_name == "continue" else get_end_do(end_name)
    do_code = """{name}do {label}{comma}variable = 1, {term}, 1
  write (6, '(I0)') variable
{endlabel} {end}
""".format(
        name=name_snippet or "",
        label=label_snippet or "",
        comma=comma_snippet or "",
        term=terminal_expression,
        endlabel=end_label or "",
        end=end_snippet,
    )
    do_expected = """  {name}DO {label}variable = 1, {term}, 1
    WRITE (6, '(I0)') variable
{endlabel} {endstmt}
""".format(
        name=name_snippet or "",
        label=label_snippet or "",
        term=terminal_expression,
        endlabel=end_label or " ",
        endstmt=get_end_do(end_name),
    )
    do_reader = FortranStringReader(do_code)
    do_reader.set_format(FortranFormat(True, False))
    do_parser = FortranParser(do_reader)
    if (name != end_name) or (label and (label != end_label)):
        with pytest.raises(AnalyzeError):
            do_parser.parse()
    else:
        do_parser.parse()
        loop = do_parser.block.content[0]
        assert str(loop).splitlines() == do_expected.splitlines()


@pytest.mark.parametrize("name", [None, "loop_name"])
@pytest.mark.parametrize("label", [None, "123"])
@pytest.mark.parametrize("control_comma", [False, True])
@pytest.mark.parametrize(
    "terminal_expression", ["1", "x+y", "size(array)", "size(this%array)"]
)
@pytest.mark.parametrize("end_name", [None, "loop_name", "wrong_name"])
@pytest.mark.parametrize("end_label", [None, "123", "456"])
def test_do_while(name, label, control_comma, terminal_expression, end_name, end_label):
    # pylint: disable=redefined-outer-name, too-many-arguments
    """
    Checks that the "do" loop parser understands the "do-while" variant of
    the syntax. This is defined in BS ISO/IEC 1539-1:2010 with R814-R822.
    """
    name_snippet = name + ": " if name else None
    label_snippet = label + " " if label else None
    comma_snippet = ", " if control_comma else None
    code = """{name}do {label}{comma}while ({term})
  write (6, '(I0)') variable
{endlabel} {endstmt}
""".format(
        name=name_snippet or "",
        label=label_snippet or "",
        comma=comma_snippet or "",
        term=terminal_expression,
        endlabel=end_label or "",
        endstmt=get_end_do(end_name),
    )
    expected = """  {name}DO {label}while ({term})
    WRITE (6, '(I0)') variable
{endlabel} {endstmt}
""".format(
        name=name_snippet or "",
        label=label_snippet or "",
        term=terminal_expression,
        endlabel=end_label or " ",
        endstmt=get_end_do(end_name),
    )
    print(code)
    reader = FortranStringReader(code)
    reader.set_format(FortranFormat(True, False))
    parser = FortranParser(reader)
    if (name != end_name) or (label and (label != end_label)):
        with pytest.raises(AnalyzeError):
            parser.parse()
    else:
        parser.parse()
        loop = parser.block.content[0]
        assert str(loop).splitlines() == expected.splitlines()



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Submodule_Stmt rule R1117
"""

from fparser.two.Fortran2008.parent_identifier_r1118 import Parent_Identifier
from fparser.two.utils import Base, ScopingRegionMixin


class Submodule_Stmt(Base, ScopingRegionMixin):  # R1117
    """
    Fortran 2008 rule R1117::

        submodule-stmt is SUBMODULE ( parent-identifier ) submodule-name

    """

    subclass_names = []
    use_names = ["Submodule_Name", "Parent_Identifier"]

    @staticmethod
    def match(fstring):
        """Check whether the input matches the rule

        :param st fstring: contains the Fortran that we are trying to match.

        :returns: instances of the Classes that have matched if there is a \
                  match or `None` if there is no match.
        :rtype: Optional[Tuple[:py:class:`fparser.two.Fortran2008.Parent_Identifier`, \
                               :py:class:`fparser.two.Fortran2008.Submodule_Name`]]

        """
        # Avoid circular dependencies by importing here.
        # pylint: disable=import-outside-toplevel
        from fparser.two.Fortran2008 import Submodule_Name

        # Import here as monkeypatching in tests fails if imported
        # from outer level.
        from fparser.common.splitline import splitparen

        # First look for "SUBMODULE"
        name = "SUBMODULE"
        if fstring[: len(name)].upper() != name:
            # the string does not start with SUBMODULE so does not
            # match
            return None
        # "SUBMODULE is found so strip it out and split the remaining
        # line by parenthesis
        splitline = splitparen(fstring[len(name) :].lstrip())
        # We expect 2 entries, the first being parent_identifier with
        # brackets and the second submodule_name. However for some
        # reason we get an additional empty 1st entry when using
        # splitline
        if len(splitline) != 3:
            # format is not ( parent_identifier ) submodule_name
            return None
        spurious = splitline[0]
        parent_id_brackets = splitline[1]
        submodule_name = splitline[2]
        if spurious:
            # this should be empty
            return None
        # Make sure the parent identifier contained in splitline[1] is
        # enclosed with round brackets
        if parent_id_brackets[0] != "(":
            return None
        if parent_id_brackets[-1] != ")":
            return None
        parent_id = parent_id_brackets[1:-1].lstrip().rstrip()
        # Format is OK from this Class' perspective. Pass on
        # parent_identifier and submodule name to appropriate classes
        return Parent_Identifier(parent_id), Submodule_Name(submodule_name)

    def tostr(self):
        """return the fortran representation of this object"""
        return f"SUBMODULE ({self.items[0]}) {self.items[1]}"

    def get_name(self):  # C1114
        """Fortran 2008 constraint C1114
        return the submodule name. This is used by the base class to check
        whether the submodule name matches the name used for the end
        submodule statement if one is provided.

        :return: the name of the submodule stored in a Name class
        :rtype: :py:class:`fparser.two.Fortran2003.Name`
        """
        return self.items[1]



# Copyright (c) 2022 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2008 rule R627

    alloc-opt is ERRMSG = errmsg-variable
              or MOLD = source-expr
              or SOURCE = source-expr
              or STAT = stat-variable

The only difference to 2003 is the addition of 'MOLD' so that's what we
test for.

"""

import pytest
from fparser.two.Fortran2008 import Allocate_Stmt, Alloc_Opt, Alloc_Opt_List


@pytest.mark.usefixtures("f2008_create")
def test_alloc_opt():
    """Test that Alloc_Opt supports the F2008 addition of 'MOLD'."""
    obj = Alloc_Opt("MOLD=c")
    assert isinstance(obj, Alloc_Opt), repr(obj)
    assert str(obj) == "MOLD = c"


@pytest.mark.usefixtures("f2008_create")
def test_allocate_stmt():
    """Check that the Fortran2008 version of allocate has picked up the
    version of Alloc_Opt that supports MOLD."""
    obj = Allocate_Stmt("allocate(b, mold=c)")
    assert obj.alloc_opt_list() == Alloc_Opt_List
    assert isinstance(obj, Allocate_Stmt), repr(obj)
    assert str(obj) == "ALLOCATE(b, MOLD = c)"
    obj = Allocate_Stmt("allocate(b, mold=c, stat=ierr)")
    assert isinstance(obj, Allocate_Stmt), repr(obj)
    assert str(obj) == "ALLOCATE(b, MOLD = c, STAT = ierr)"



# Copyright (c) 2024 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2008 rule R856

error-stop-stmt is ERROR STOP [ stop-code ]
"""

import pytest

from fparser.api import get_reader
from fparser.two.utils import NoMatchError, walk
from fparser.two import Fortran2003, utils
from fparser.two.Fortran2008 import Stop_Code


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize("string", ["1", "- 1", '"abc"', "'abc'", "'abc' // 'def'"])
def test_simple_stop_code(string):
    """Test that error-stop matches the expected valid values."""
    result = Stop_Code(string)
    assert str(result) == string


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize("string", ["call sub()", "do i", "1, 2, 3"])
def test_simple_stop_code_errors(string):
    """Test that invalid stop codes are handled."""
    with pytest.raises(NoMatchError) as err:
        Stop_Code(string)
    assert f"Stop_Code: '{string}'" in str(err.value)


@pytest.mark.parametrize("string", ["1", "12345"])
def test_stop_stmt_2003_stop_code(f2008_parser, string):
    """Test that 'stop' parsing works in real code, and returns a 2003
    StopCode. This is the case if the stop code is between one and
    five digits only:
    """
    code = f"""
    subroutine dummy()
    stop {string}
    end subroutine dummy
    """
    tree = f2008_parser(get_reader(code))
    stop_code = walk(tree, Fortran2003.Stop_Code)[0]
    assert str(stop_code) == string


@pytest.mark.parametrize("string", ["1234567", "12 .AND. 34"])
def test_stop_stmt_2008(f2008_parser, string, monkeypatch):
    """Test that stop parsing works in real code when using F2008
    only (i.e. not F2003) statements. Note that '12 .and. 34' is a
    level-5-expr, and as such would not be accepted by the F2003
    "extended-stop-args" extension in fparser.
    """
    monkeypatch.setattr(utils, "_EXTENSIONS", [])
    code = f"""
    subroutine dummy()
    stop {string}
    end subroutine dummy
    """
    tree = f2008_parser(get_reader(code))
    stop_stmt = walk(tree, Fortran2003.Stop_Stmt)[0]
    assert str(stop_stmt.children[1]) == string



# Copyright (c) 2021 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R706 : This file tests support for the
Level_2_Expr class.

"""

import pytest
from fparser.two.Fortran2003 import Level_2_Expr
from fparser.two.parser import ParserFactory

# This is required to setup the fortran2003 classes (when matching
# with Level_2_Expr directly)
_ = ParserFactory().create(std="f2003")


@pytest.mark.parametrize(
    "string,str_repr",
    [
        (
            "a + c ** b",
            "Level_2_Expr(Name('a'), '+', Mult_Operand(Name('c'), '**'," " Name('b')))",
        ),
        (
            "a + c * b",
            "Level_2_Expr(Name('a'), '+', Add_Operand(Name('c'), '*'," " Name('b')))",
        ),
        (
            "a + 1.0E-10 * b",
            "Level_2_Expr(Name('a'), '+', Add_Operand("
            "Real_Literal_Constant('1.0E-10', None), '*', Name('b')))",
        ),
        (
            "a + b * 1.0E-10",
            "Level_2_Expr(Name('a'), '+', Add_Operand("
            "Name('b'), '*', Real_Literal_Constant('1.0E-10', None)))",
        ),
        (
            "a + 1.0d+10 * b",
            "Level_2_Expr(Name('a'), '+', Add_Operand("
            "Real_Literal_Constant('1.0D+10', None), '*', Name('b')))",
        ),
        (
            "- .0E-10 * b - a",
            "Level_2_Expr(Level_2_Unary_Expr('-', Add_Operand("
            "Real_Literal_Constant('.0E-10', None), '*', Name('b'))), '-', "
            "Name('a'))",
        ),
    ],
)
def test_level2_exprn(string, str_repr):
    """Test for a successful match with a valid level-2 expression."""
    result = Level_2_Expr(string)
    assert str(result).lower() == string.lower()
    assert repr(result) == str_repr



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Explicit_Coshape_Spec rule R511
"""

from fparser.common.splitline import string_replace_map
from fparser.two.Fortran2008.lower_cobound_r512 import Lower_Cobound
from fparser.two.utils import SeparatorBase


class Explicit_Coshape_Spec(SeparatorBase):  # R511
    """
    Fortran 2008 rule R511
    explicit-coshape-spec is [ coshape-spec-list , ] [ lower-cobound : ] *

    Associated constraint is:

    "C529 (R511)  A lower-cobound or upper-cobound that  is  not  a  constant
          expression shall appear only in a subprogram, BLOCK construct, or
          interface body."

    C529 is currently not checked - issue #259.

    """

    subclass_names = []
    use_names = ["Coshape_Spec_List", "Lower_Cobound"]

    @staticmethod
    def match(string):
        """
        Implements the matching for explicit coarray shape specification.

        :param str string: the string to match as deferred shape.

        :return: `None` if there is no match, otherwise a 2-tuple \
                 containing matched coshape-spec-list or `None` and \
                 matched lower-cobound or `None`.
        :rtype: `NoneType` or \
            (:py:class:`fparser.two.Fortran2008.Coshape_Spec_List` or `None`, \
             :py:class:`fparser.two:Fortran2008.Lower_Cobound` or `None`)

        """
        # Avoid circular dependencies by importing here.
        # pylint: disable=import-outside-toplevel
        from fparser.two.Fortran2008 import Coshape_Spec_List

        if not string.endswith("*"):
            return None
        line = string[:-1].rstrip()
        if not line:
            return (None, None)
        if line.endswith(":"):
            line, repmap = string_replace_map(line[:-1].rstrip())
            sep_pos = line.rfind(",")
            if sep_pos == -1:
                return (None, Lower_Cobound(repmap(line)))
            return (
                Coshape_Spec_List(repmap(line[:sep_pos].rstrip())),
                Lower_Cobound(repmap(line[sep_pos + 1 :].lstrip())),
            )
        if not line.endswith(","):
            return None
        line = line[:-1].rstrip()
        return (Coshape_Spec_List(line), None)

    def tostr(self):
        """
        Converts the explicit coarray shape specification to string.

        :return: the shape specification as string.
        :rtype: str

        """
        s = ""
        if self.items[0]:
            s += str(self.items[0]) + ", "
        if self.items[1]:
            s += str(self.items[1]) + " : "
        s += "*"
        return s



# Copyright (c) 2018-2022 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R1109 : This file tests the support for the
Use statement.

"""

import pytest
from fparser.api import get_reader
from fparser.two.Fortran2003 import Use_Stmt
from fparser.two.symbol_table import SYMBOL_TABLES, ModuleUse
from fparser.two.utils import NoMatchError, InternalError


@pytest.fixture(autouse=True)
def use_f2003(f2003_create):
    """
    A pytest fixture that just sets things up so that the f2003_create
    fixture is used automatically for every test in this file.
    """


# match() use ...


# match() 'use x'. Use both string and reader input here, but from
# here on we will just use string input as that is what is passed to
# the match() method
def test_use():
    """Check that a basic use is parsed correctly. Input separately as a
    string and as a reader object

    """

    def check_use(reader):
        """Internal helper function to avoid code replication."""
        ast = Use_Stmt(reader)
        assert "USE my_model" in str(ast)
        assert repr(ast) == "Use_Stmt(None, None, Name('my_model'), '', None)"

    line = "use my_model"
    check_use(line)
    reader = get_reader(line)
    check_use(reader)


# match() 'use :: x'
def test_use_colons():
    """Check that a basic use with '::' is parsed correctly."""
    line = "use :: my_model"
    ast = Use_Stmt(line)
    assert "USE :: my_model" in str(ast)
    assert repr(ast) == "Use_Stmt(None, '::', Name('my_model'), '', None)"


# match() 'use, nature :: x'
def test_use_nature():
    """Check that a use with a 'nature' specification is parsed correctly."""
    line = "use, intrinsic :: my_model"
    ast = Use_Stmt(line)
    assert "USE, INTRINSIC :: my_model" in str(ast)
    assert repr(ast) == (
        "Use_Stmt(Module_Nature('INTRINSIC'), '::', Name('my_model'), " "'', None)"
    )


# match() 'use x, rename'
@pytest.mark.usefixtures("fake_symbol_table")
def test_use_rename():
    """Check that a use with a rename clause is parsed correctly."""
    line = "use my_module, new_name=>name"
    ast = Use_Stmt(line)
    assert "USE my_module, new_name => name" in str(ast)
    assert repr(ast) == (
        "Use_Stmt(None, None, Name('my_module'), ',', Rename_List(',', "
        "(Rename(None, Name('new_name'), Name('name')),)))"
    )
    table = SYMBOL_TABLES.current_scope
    assert "my_module" in table._modules
    use = table._modules["my_module"]
    # We only store the *local* names of the symbols in rename_list
    assert use.rename_list == ["new_name"]
    # The original name of the symbol in the module being imported is stored
    # in a separate map.
    assert use._local_to_module_map["new_name"] == "name"


def test_use_operator_rename():
    """
    Check that a use with a rename clause for an operator is parsed correctly.

    """
    line = "use my_module, operator(.new.) => operator(.old.)"
    ast = Use_Stmt(line)
    assert repr(ast) == (
        "Use_Stmt(None, None, Name('my_module'), ',', Rename_List(',', "
        "(Rename('OPERATOR', Defined_Op('.NEW.'), Defined_Op('.OLD.')),)))"
    )
    SYMBOL_TABLES.enter_scope("test_scope")
    ast = Use_Stmt(line)
    # Check that the symbol table functionality copes (by ignoring operators).
    table = SYMBOL_TABLES.current_scope
    use = table._modules["my_module"]
    # Operators are not currently captured in the SymbolTable (TODO #379)
    assert use.rename_list is None
    SYMBOL_TABLES.exit_scope()


# match() 'use x, only: y'
def test_use_only():
    """Check that a use statement is parsed correctly when there is an
    only clause. Test both with and without a scoping region.

    """
    line = "use my_model, only: name"
    ast = Use_Stmt(line)
    assert "USE my_model, ONLY: name" in str(ast)
    assert repr(ast) == (
        "Use_Stmt(None, None, Name('my_model'), ', ONLY:', Only_List(',', "
        "(Name('name'),)))"
    )
    # Repeat when there is a scoping region.
    SYMBOL_TABLES.enter_scope("test_scope")
    ast = Use_Stmt(line)
    table = SYMBOL_TABLES.current_scope
    assert "my_model" in table._modules
    use = table._modules["my_model"]
    assert use.name == "my_model"
    assert use.only_list == ["name"]
    assert use.rename_list is None
    SYMBOL_TABLES.exit_scope()


# match() 'use x, only:'
def test_use_only_empty():
    """Check that a use statement is parsed correctly when there is an
    only clause without any content. Test both with and without a scoping region.

    """
    line = "use my_model, only:"
    ast = Use_Stmt(line)
    assert "USE my_model, ONLY:" in str(ast)
    assert repr(ast) == ("Use_Stmt(None, None, Name('my_model'), ', ONLY:', None)")
    # Repeat when there is a scoping region.
    SYMBOL_TABLES.enter_scope("test_scope")
    ast = Use_Stmt(line)
    table = SYMBOL_TABLES.current_scope
    assert "my_model" in table._modules
    use = table._modules["my_model"]
    assert isinstance(use, ModuleUse)
    assert use.only_list == []
    assert use.rename_list is None
    SYMBOL_TABLES.exit_scope()


# match() 'use x, only: b => c'
def test_use_only_plus_rename():
    """Check that a use statement with an only clause and some variable
    renaming is parsed correctly.

    """
    line = "use my_model, only: a, b=>c"
    ast = Use_Stmt(line)
    assert repr(ast) == (
        "Use_Stmt(None, None, Name('my_model'), "
        "', ONLY:', Only_List(',', (Name('a'), "
        "Rename(None, Name('b'), Name('c')))))"
    )
    # Repeat when there is a scoping region.
    SYMBOL_TABLES.enter_scope("test_scope")
    ast = Use_Stmt(line)
    table = SYMBOL_TABLES.current_scope
    assert "my_model" in table._modules
    use = table._modules["my_model"]
    assert isinstance(use, ModuleUse)
    assert use.symbol_names == ["a", "b"]
    assert sorted(use.only_list) == ["a", "b"]
    assert use.rename_list is None
    assert use._local_to_module_map["b"] == "c"
    SYMBOL_TABLES.exit_scope()


# match() 'use x, only: operator(-)'
def test_use_only_operator():
    """
    Check that a 'use, only' that imports an operator is parsed correctly.

    """
    line = "use my_mod, only: operator(-), operator(.yes.)"
    ast = Use_Stmt(line)
    assert repr(ast) == (
        "Use_Stmt(None, None, Name('my_mod'), ', ONLY:', Only_List(',', "
        "(Generic_Spec('OPERATOR', Extended_Intrinsic_Op('-')),"
        " Generic_Spec('OPERATOR', Defined_Op('.YES.')))))"
    )
    # Repeat when there is a scoping region.
    SYMBOL_TABLES.enter_scope("test_scope")
    ast = Use_Stmt(line)
    table = SYMBOL_TABLES.current_scope
    use = table._modules["my_mod"]
    # Operators are not currently captured in the symbol table.
    # TODO #379.
    assert use.only_list == []
    SYMBOL_TABLES.exit_scope()


@pytest.mark.parametrize("lhs, rhs", [("-", "+"), (".in.", ".out.")])
def test_use_only_renamed_operator(lhs, rhs):
    """
    Check that a 'use, only' that imports and locally renames an operator is
    correctly parsed. We test both for an extended intrinsic operator and a
    defined operator.

    """
    line = f"use my_mod, only: operator({lhs}) => operator({rhs})"
    ast = Use_Stmt(line)
    if lhs == "-":
        assert repr(ast) == (
            "Use_Stmt(None, None, Name('my_mod'), ', ONLY:', Only_List(',', "
            "(Generic_Spec('OPERATOR', Extended_Intrinsic_Op('-) => "
            "operator(+')),)))"
        )
    else:
        assert repr(ast) == (
            "Use_Stmt(None, None, Name('my_mod'), ', ONLY:', Only_List(',', "
            "(Rename('OPERATOR', Defined_Op('.IN.'), Defined_Op('.OUT.')),)))"
        )

    # Repeat when there is a scoping region.
    SYMBOL_TABLES.enter_scope("test_scope")
    ast = Use_Stmt(line)
    table = SYMBOL_TABLES.current_scope
    use = table._modules["my_mod"]
    # Operators are not currently captured in the symbol table.
    # TODO #379.
    assert use.only_list == []
    SYMBOL_TABLES.exit_scope()


# match() '  use  ,  nature  ::  x  ,  name=>new_name'
def test_use_spaces_1():
    """Check that a use statement with spaces works correctly with
    renaming.

    """
    line = "  Use  ,  intrinsic  ::  my_model  ,  name=>new_name  "
    ast = Use_Stmt(line)
    assert "USE, INTRINSIC :: my_model, name => new_name" in str(ast)
    assert repr(ast) == (
        "Use_Stmt(Module_Nature('INTRINSIC'), '::', Name('my_model'), ',', "
        "Rename_List(',', (Rename(None, Name('name'), Name('new_name')),)))"
    )


# match() '  use  ,  nature  ::  x  ,  only  :  name'
def test_use_spaces_2():
    """Check that a use statement with spaces works correctly with an only
    clause.

    """
    line = "  use  ,  intrinsic  ::  my_model  ,  only  :  name  "
    ast = Use_Stmt(line)
    assert "USE, INTRINSIC :: my_model, ONLY: name" in str(ast)
    assert (
        repr(ast) == "Use_Stmt(Module_Nature('INTRINSIC'), '::', Name('my_model'), ', "
        "ONLY:', Only_List(',', (Name('name'),)))"
    )


# match() mixed case
def test_use_mixed_case():
    """Check that a use statement with mixed case keywords ('use' and
    'only') works as expected.

    """
    line = "UsE my_model, OnLy: name"
    ast = Use_Stmt(line)
    assert "USE my_model, ONLY: name" in str(ast)
    assert (
        repr(ast) == "Use_Stmt(None, None, Name('my_model'), ', ONLY:', Only_List(',', "
        "(Name('name'),)))"
    )


# match() Syntax errors


def test_syntaxerror():
    """Test that NoMatchError is raised for various syntax errors."""
    for line in [
        "us",
        "ust",
        "use",
        "usemy_model",
        "use, ",
        "use, ::",
        "use, intrinsic",
        "use, intrinsic::",
        "use, intrinsic my_module",
        "use,",
        "use my_model,",
        "use my_model, only",
        "use my_model, only ;",
        "use my_model, only name",
    ]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Use_Stmt(line)
        assert f"Use_Stmt: '{line}'" in str(excinfo.value)


# match() Internal errors


def test_use_internal_error1():
    """Check that an internal error is raised if the length of the Items
    list is not 5 as the str() method assumes that it is.

    """
    line = "use my_model"
    ast = Use_Stmt(line)
    ast.items = (None, None, None, None)
    with pytest.raises(InternalError) as excinfo:
        str(ast)
    assert "should be of size 5 but found '4'" in str(excinfo.value)


def test_use_internal_error2():
    """Check that an internal error is raised if the module name (entry 2
    of Items) is empty or None as the str() method assumes that it is
    a string with content.

    """
    line = "use my_model"
    ast = Use_Stmt(line)
    for content in [None, ""]:
        ast.items = (None, None, content, None, None)
        with pytest.raises(InternalError) as excinfo:
            str(ast)
        assert "entry 2 should be a module name but it is empty" in str(excinfo.value)


def test_use_internal_error3():
    """Check that an internal error is raised if entry 3 of Items is
    'None' as the str() method assumes it is a (potentially empty)
    string.

    """
    line = "use my_model"
    ast = Use_Stmt(line)
    ast.items = (None, None, "my_module", None, None)
    with pytest.raises(InternalError) as excinfo:
        str(ast)
    assert "entry 3 should be a string but found 'None'" in str(excinfo.value)


@pytest.mark.usefixtures("fake_symbol_table")
def test_use_internal_error_only_list(monkeypatch):
    """Check that an internal error is raised if an Only_List contains an
    unexpected entry.

    """
    # First get a valid parse tree and then break it.
    line = "use my_model, only: var"
    ast = Use_Stmt(line)
    only_list = ast.children[4]
    monkeypatch.setattr(only_list, "items", ("wrong",))
    # Monkeypatch the underlying _match method to return this broken tree.
    monkeypatch.setattr(Use_Stmt, "_match", lambda x: ast.children)
    with pytest.raises(InternalError) as err:
        Use_Stmt(line)
    assert (
        "An Only_List can contain only Name, Rename or Generic_Spec entries but "
        "found 'str' when matching 'use my_model, only: var'" in str(err.value)
    )



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2026, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran 2008 Format_Item rule R1003
"""

from typing import Optional, Tuple, Union

from fparser.two.Fortran2003 import (
    Digit_String,
    Format_Item as Format_Item_2003,
    Format_Item_List,
)
from fparser.two.utils import NoMatchError


class Format_Item(Format_Item_2003):  # R1003
    """Fortran 2008 rule R1003

    format-item is [ r ] data-edit-desc
                or control-edit-desc
                or char-string-edit-desc
                or [ r ] ( format-item-list )
                or format-item-c1002
                or hollerith-item
                or * ( format-item-list )

    Extends the Fortran 2003 rule R1003 with the additional unlimited
    format repeat specifier ``*`` before a parenthesised format-item-list.

    """

    # Inherit the parent's subclass_names so that Control_Edit_Desc,
    # Hollerith_Item, etc. remain registered when the F2008 parser
    # replaces F2003 Format_Item in the class hierarchy.
    # Include "Format_Item" itself so that this F2008 class is
    # registered in Base.subclasses["Format_Item"] and can be
    # discovered when F2003 code (e.g. Format_Item_List.match)
    # directly references the F2003 Format_Item class.
    subclass_names = Format_Item_2003.subclass_names[:] + ["Format_Item"]
    use_names = Format_Item_2003.use_names[:]

    @staticmethod
    def match(
        string: str,
    ) -> Optional[
        Tuple[Union[str, Digit_String], Union[Format_Item_2003, Format_Item_List]]
    ]:
        """Attempts to match the supplied text with this rule.

        Calls the Fortran 2003 match first. If that fails, checks
        for the Fortran 2008 unlimited format repeat: ``*(format-item-list)``.

        :param string: Fortran code to check for a match.

        :returns: None if there is no match, a tuple of size 2
            containing a repeat specifier (``"*"`` or a Repeat instance)
            and the matched descriptor or format-item-list.

        """
        if not string:
            return None
        # Fortran 2003 matches all but unlimited repeat, so try it first.
        # The F2003 match raises NoMatchError rather than returning None
        # when it cannot parse the string (e.g. Data_Edit_Desc fails).
        try:
            result = Format_Item_2003.match(string)
        except NoMatchError:
            result = None
        if result:
            return result
        # Try to match unlimited format repeat: *(format-item-list)
        strip_string = string.strip()
        if not strip_string:
            return None
        if strip_string[0] == "*" and len(strip_string) > 1:
            my_string = strip_string[1:].lstrip()
            if my_string[0] == "(" and my_string[-1] == ")":
                return ("*", Format_Item_List(my_string[1:-1].lstrip()))
        return None



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Implicit_Part_Stmt constraint C1112
"""

from fparser.two.Fortran2003 import Implicit_Part
from fparser.two.Fortran2008.implicit_part_stmt_c1112 import Implicit_Part_Stmt_C1112
from fparser.two.utils import BlockBase


class Implicit_Part_C1112(Implicit_Part):  # C1112
    """Fortran 2008 constraint C1112
    C1112 A submodule specification-part shall not contain a
    format-stmt, entry-stmt, or stmt-function-stmt.

    This class specialises 'Implicit_Part' so that the specialised
    'Implicit_Part_Stmt_C1112' is called rather than the original
    'Implicit_Part_Stmt'

    """

    use_names = ["Implicit_Part_Stmt_C1112", "Implicit_Stmt"]

    @staticmethod
    def match(reader):
        """Check whether the input matches the rule

        param reader: the fortran file reader containing the line(s)
                      of code that we are trying to match
        :type reader: :py:class:`fparser.common.readfortran.FortranFileReader`
                      or
                      :py:class:`fparser.common.readfortran.FortranStringReader`
        :return: `tuple` containing a single `list` which contains
                 instance of the classes that have matched if there is
                 a match or `None` if there is no match

        """
        return BlockBase.match(None, [Implicit_Part_Stmt_C1112], None, reader)



# Copyright (c) 2022 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2008 rule R214

action-stmt is allocate-stmt
                or assignment-stmt
                or backspace-stmt
                or call-stmt
                or close-stmt
                or continue-stmt
                or cycle-stmt
                or deallocate-stmt
                or end-function-stmt
                or end-mp-subprogram-stmt
                or end-program-stmt
                or end-subroutine-stmt
                or endfile-stmt
                or error-stop-stmt
                or exit-stmt
                or flush-stmt
                or forall-stmt
                or goto-stmt
                or if-stmt
                or inquire-stmt
                or lock-stmt
                or nullify-stmt
                or open-stmt
                or pointer-assignment-stmt
                or print-stmt
                or read-stmt
                or return-stmt
                or rewind-stmt
                or stop-stmt
                or sync-all-stmt
                or sync-images-stmt
                or sync-memory-stmt
                or unlock-stmt
                or wait-stmt
                or where-stmt
                or write-stmt
                or arithmetic-if-stmt
                or computed-goto-stmt

Associated constraints are:

"C201 (R208) An execution-part shall not contain an end-function-stmt,
      end-mp-subprogram-stmt, end-program-stmt, or end-subroutine-stmt."

TODO: Implement and test lock-stmt, sync-all-stmt, sync-images-stmt,
      sync-memory-stmt, unlock-stmt (#321)
"""

import pytest
from fparser.api import get_reader, walk
from fparser.two.Fortran2003 import (
    Allocate_Stmt,
    Arithmetic_If_Stmt,
    Backspace_Stmt,
    Call_Stmt,
    Close_Stmt,
    Computed_Goto_Stmt,
    Continue_Stmt,
    Cycle_Stmt,
    Deallocate_Stmt,
    End_Function_Stmt,
    End_Subroutine_Stmt,
    Endfile_Stmt,
    Exit_Stmt,
    Flush_Stmt,
    Forall_Stmt,
    Goto_Stmt,
    If_Stmt,
    Inquire_Stmt,
    Nullify_Stmt,
    Open_Stmt,
    Pointer_Assignment_Stmt,
    Print_Stmt,
    Read_Stmt,
    Return_Stmt,
    Rewind_Stmt,
    Stop_Stmt,
    Wait_Stmt,
    Where_Stmt,
    Write_Stmt,
)
from fparser.two.Fortran2008 import Action_Stmt, Error_Stop_Stmt
from fparser.two.utils import NoMatchError


@pytest.mark.usefixtures("f2008_create", "fake_symbol_table")
@pytest.mark.parametrize(
    "string, cls",
    [
        ("ALLOCATE(A(10))", Allocate_Stmt),
        ("BACKSPACE 42", Backspace_Stmt),
        ("CALL SOME_ROUTINE()", Call_Stmt),
        ("CLOSE(23)", Close_Stmt),
        ("CONTINUE", Continue_Stmt),
        ("CYCLE", Cycle_Stmt),
        ("DEALLOCATE(A)", Deallocate_Stmt),
        ("END FUNCTION FUNC", End_Function_Stmt),
        ("END SUBROUTINE MYSUB", End_Subroutine_Stmt),
        ("ENDFILE 42", Endfile_Stmt),
        ("EXIT", Exit_Stmt),
        ("FLUSH 23", Flush_Stmt),
        ("FORALL (I=1:N) A(I,I) = X(I)", Forall_Stmt),
        ("GO TO 915", Goto_Stmt),
        ("IF (A > B) A = B", If_Stmt),
        ("INQUIRE (IOLENGTH=N) M", Inquire_Stmt),
        ("NULLIFY(P)", Nullify_Stmt),
        ("OPEN (10, FILE = 'employee.names', ACTION = 'READ', PAD = 'YES')", Open_Stmt),
        ("P => NULL()", Pointer_Assignment_Stmt),
        ("PRINT *, 'HELLO WORLD'", Print_Stmt),
        ("READ (6, *) VAR", Read_Stmt),
        ("RETURN", Return_Stmt),
        ("REWIND 23", Rewind_Stmt),
        ("STOP", Stop_Stmt),
        ("WAIT(23)", Wait_Stmt),
        ("WHERE (TEMP > 100.0) TEMP = TEMP - REDUCE_TEMP", Where_Stmt),
        ("WRITE(10, *) A", Write_Stmt),
        ("IF (A - B) 1, 2, 3", Arithmetic_If_Stmt),
        ("GO TO (1, 2, 3) 2-1", Computed_Goto_Stmt),
    ],
)
def test_other(string, cls):
    """Test that previous subclasses are still matched correctly."""
    result = Action_Stmt(string)
    assert isinstance(result, cls)


def test_other_functional(f2008_parser):
    """Test previous subclasses are still matched correctly in a subroutine."""
    tree = f2008_parser(get_reader("""\
subroutine my_abort
stop
end subroutine my_abort
    """))
    assert walk(tree, Stop_Stmt)
    assert "STOP" in str(tree)


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize("string, cls", [("ERROR STOP", Error_Stop_Stmt)])
def test_new_f2008_stmt(string, cls):
    """Test that newly added F2008 subclasses are matched correctly."""
    result = Action_Stmt(string)
    assert isinstance(result, cls)


@pytest.mark.usefixtures("f2008_create")
def test_error():
    """Test that invalid syntax raises an error."""
    with pytest.raises(NoMatchError) as excinfo:
        _ = Action_Stmt("ERR STOP")
    assert "Action_Stmt: 'ERR STOP'" in str(excinfo.value)



# Copyright (c) 2022 Science and Technology Facilities Council
#
# All rights reserved.
##
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Tests for the fparser2_bench script."""

import pytest
from fparser.scripts import fparser2_bench


def test_gen_sub():
    """Check the gen_sub() routine works as expected."""
    code = fparser2_bench.gen_sub(3)
    assert "subroutine g3(x)\n" in code
    assert "do i = 3, 12\n" in code
    assert "end subroutine\n" in code


def test_create_bench():
    """Check the create_bench() routine works as expected."""
    code = fparser2_bench.create_bench(3)
    assert "program bench3\n" in code
    assert code.count("end subroutine") == 3
    for idx in range(1, 4):
        assert f"subroutine g{idx}(x)" in code
    assert "end program" in code


def test_runner_invalid_num_routines():
    """Test the checking on the value of the supplied num_routines
    parameter."""
    with pytest.raises(ValueError) as err:
        fparser2_bench.runner(0)
    assert (
        "Number of routines to create must be a positive, non-zero "
        "integer but got: 0" in str(err.value)
    )


def test_runner(capsys):
    """Check that normal usage gives the expected benchmark output."""
    fparser2_bench.runner(3)
    stdout, stderr = capsys.readouterr()
    assert stderr == ""
    assert "Constructing benchmark code with 3 subroutines" in stdout
    assert "Time taken for parse =" in stdout



# Modified work Copyright (c) 2017-2022 Science and Technology
# Facilities Council.
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Tools for constructing patterns.

Permission to use, modify, and distribute this software is given under the
terms of the NumPy License. See http://scipy.org.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Author: Pearu Peterson <pearu@cens.ioc.ee>
Created: Oct 2006

"""

import re

dollar_ok = True


class Pattern:
    """

    ::

        p1 | p2    -> <p1> | <p2>
        p1 + p2    -> <p1> <p2>
        p1 & p2    -> <p1><p2>
        ~p1        -> [ <p1> ]
        ~~p1       -> [ <p1> ]...
        ~~~p1      -> <p1> [ <p1> ]...
        ~~~~p1     -> ~~~p1
        abs(p1)    -> whole string match of <p1>
        p1.named(name) -> match of <p1> has name
        p1.match(string) -> return string match with <p1>
        p1.flags(<re.I,..>)
        p1.rsplit(..) -> split a string from the rightmost p1 occurrence
        p1.lsplit(..) -> split a string from the leftmost p1 occurrence

    """

    _special_symbol_map = {
        ".": "[.]",
        "*": "[*]",
        "+": "[+]",
        "|": "[|]",
        "(": r"\(",
        ")": r"\)",
        "[": r"\[",
        "]": r"\]",
        "^": "[^]",
        "$": "[$]",
        "?": "[?]",
        "{": r"\{",
        "}": r"\}",
        ">": "[>]",
        "<": "[<]",
        "=": "[=]",
    }

    def __init__(self, label, pattern, optional=0, flags=0, value=None):
        self.label = label
        self.pattern = pattern
        self.optional = optional
        self._flags = flags
        self.value = value

    def flags(self, *flags):
        f = self._flags
        for f1 in flags:
            f = f | f1
        return Pattern(
            self.label, self.pattern, optional=self.optional, flags=f, value=self.value
        )

    def get_compiled(self):
        try:
            return self._compiled_pattern
        except AttributeError:
            self._compiled_pattern = compiled = re.compile(self.pattern, self._flags)
            return compiled

    def match(self, string):
        return self.get_compiled().match(string)

    def search(self, string):
        return self.get_compiled().search(string)

    def rsplit(self, string, is_add=False):
        """
        Return (<lhs>, <pattern_match>, <rhs>) where::

            string = lhs + pattern_match + rhs

        and rhs does not contain pattern_match.
        If no pattern_match is found in string, return None.
        """
        compiled = self.get_compiled()
        t = compiled.split(string)
        if is_add:
            n = "".join(t[-3:]).replace(" ", "")
            if abs_real_literal_constant.match(n):
                t = t[:-3] + [n]
        if len(t) < 3:
            return
        if "" in t[1:-1]:
            return
        rhs = t[-1].strip()
        pattern_match = t[-2].strip()
        assert abs(self).match(pattern_match), repr((self, string, t, pattern_match))
        lhs = ("".join(t[:-2])).strip()
        return lhs, pattern_match, rhs

    def lsplit(self, string):
        """
        Return (<lhs>, <pattern_match>, <rhs>) where::

            string = lhs + pattern_match + rhs

        and rhs does not contain pattern_match.
        If no pattern_match is found in string, return None.
        """
        compiled = self.get_compiled()
        t = compiled.split(string)  # can be optimized
        if len(t) < 3:
            return
        lhs = t[0].strip()
        pattern_match = t[1].strip()
        rhs = ("".join(t[2:])).strip()
        assert abs(self).match(pattern_match), repr(pattern_match)
        return lhs, pattern_match, rhs

    def __abs__(self):
        return Pattern(
            self.label,
            r"\A" + self.pattern + r"\Z",
            flags=self._flags,
            value=self.value,
        )

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.label, self.pattern)

    def __or__(self, other):
        label = "( %s OR %s )" % (self.label, other.label)
        if self.pattern == other.pattern:
            pattern = self.pattern
            flags = self._flags
        else:
            pattern = "(%s|%s)" % (self.pattern, other.pattern)
            flags = self._flags | other._flags
        return Pattern(label, pattern, flags=flags)

    def __and__(self, other):
        if isinstance(other, Pattern):
            label = "%s%s" % (self.label, other.label)
            pattern = self.pattern + other.pattern
            flags = self._flags | other._flags
        else:
            assert isinstance(other, str), repr(other)
            label = "%s%s" % (self.label, other)
            pattern = self.pattern + other
            flags = self._flags
        return Pattern(label, pattern, flags=flags)

    def __rand__(self, other):
        assert isinstance(other, str), repr(other)
        label = "%s%s" % (other, self.label)
        pattern = other + self.pattern
        return Pattern(label, pattern, flags=self._flags)

    def __invert__(self):
        if self.optional:
            if self.optional == 1:
                return Pattern(
                    self.label + "...",
                    self.pattern[:-1] + "*",
                    optional=2,
                    flags=self._flags,
                )
            if self.optional == 2:
                return Pattern(
                    "%s %s" % (self.label[1:-4].strip(), self.label),
                    self.pattern[:-1] + "+",
                    optional=3,
                    flags=self._flags,
                )
            return self
        label = "[ %s ]" % (self.label)
        pattern = "(%s)?" % (self.pattern)
        return Pattern(label, pattern, optional=1, flags=self._flags)

    def __add__(self, other):
        if isinstance(other, Pattern):
            label = "%s %s" % (self.label, other.label)
            pattern = self.pattern + r"\s*" + other.pattern
            flags = self._flags | other._flags
        else:
            assert isinstance(other, str), repr(other)
            label = "%s %s" % (self.label, other)
            other = self._special_symbol_map.get(other, other)
            pattern = self.pattern + r"\s*" + other
            flags = self._flags
        return Pattern(label, pattern, flags=flags)

    def __radd__(self, other):
        assert isinstance(other, str), repr(other)
        label = "%s %s" % (other, self.label)
        other = self._special_symbol_map.get(other, other)
        pattern = other + r"\s*" + self.pattern
        return Pattern(label, pattern, flags=self._flags)

    def named(self, name=None):
        if name is None:
            label = self.label
            assert label[0] + label[-1] == "<>" and " " not in label, repr(label)
        else:
            label = "<%s>" % (name)
        pattern = "(?P%s%s)" % (label.replace("-", "_"), self.pattern)
        return Pattern(label, pattern, flags=self._flags, value=self.value)

    def rename(self, label):
        if label[0] + label[-1] != "<>":
            label = "<%s>" % (label)
        return Pattern(
            label,
            self.pattern,
            optional=self.optional,
            flags=self._flags,
            value=self.value,
        )

    def __call__(self, string):
        m = self.match(string)
        if m is None:
            return
        if self.value is not None:
            return self.value
        return m.group()


# Predefined patterns


letter = Pattern("<letter>", "[A-Z]", flags=re.I)
if dollar_ok:
    name = Pattern("<name>", r"[A-Z][\w$]*", flags=re.I)
else:
    name = Pattern("<name>", r"[A-Z]\w*", flags=re.I)
# file_name pattern is start of match '^' to end of match '$', either
# match a single character that is not space '\S', or '|' a single
# character that is not space at the start '\S' and end '\S' of the
# match with anything '.*' inbetween.
file_name = Pattern("<file_name>", r"^(\S|\S.*\S)$", flags=re.I)
macro_name = Pattern("<macro_name>", r"[A-Z_]\w*", flags=re.I)
abs_macro_name = abs(macro_name)
digit = Pattern("<digit>", r"\d")
underscore = Pattern("<underscore>", "_")
binary_digit = Pattern("<binary-digit>", r"[01]")
octal_digit = Pattern("<octal-digit>", r"[0-7]")
hex_digit = Pattern("<hex-digit>", r"[\dA-F]", flags=re.I)

digit_string = Pattern("<digit-string>", r"\d+")
abs_digit_string = abs(digit_string)
abs_digit_string_named = abs(digit_string.named("value"))
binary_digit_string = Pattern("<binary-digit-string>", r"[01]+")
octal_digit_string = Pattern("<octal-digit-string>", r"[0-7]+")
hex_digit_string = Pattern("<hex-digit-string>", r"[\dA-F]+", flags=re.I)

sign = Pattern("<sign>", r"[+-]")
exponent_letter = Pattern("<exponent-letter>", r"[ED]", flags=re.I)

alphanumeric_character = Pattern("<alphanumeric-character>", r"\w")  # [A-Z0-9_]
special_character = Pattern(
    "<special-character>", r'[ =+-*/\()[\]{},.:;!"%&~<>?,\'`^|$#@]'
)
character = alphanumeric_character | special_character

kind_param = digit_string | name
kind_param_named = kind_param.named("kind-param")
signed_digit_string = ~sign + digit_string
int_literal_constant = digit_string + ~("_" + kind_param)
signed_int_literal_constant = ~sign + int_literal_constant
int_literal_constant_named = digit_string.named("value") + ~("_" + kind_param_named)
signed_int_literal_constant_named = (~sign + digit_string).named("value") + ~(
    "_" + kind_param_named
)

binary_constant = (
    "B" + ("'" & binary_digit_string & "'" | '"' & binary_digit_string & '"')
).flags(re.I)
octal_constant = (
    "O" + ("'" & octal_digit_string & "'" | '"' & octal_digit_string & '"')
).flags(re.I)
hex_constant = (
    "Z" + ("'" & hex_digit_string & "'" | '"' & hex_digit_string & '"')
).flags(re.I)
boz_literal_constant = binary_constant | octal_constant | hex_constant

exponent = signed_digit_string
significand = digit_string + "." + ~digit_string | "." + digit_string
real_literal_constant = significand + ~(exponent_letter + exponent) + ~(
    "_" + kind_param
) | digit_string + exponent_letter + exponent + ~("_" + kind_param)
real_literal_constant_named = (
    significand + ~(exponent_letter + exponent)
    | digit_string + exponent_letter + exponent
).named("value") + ~("_" + kind_param_named)
signed_real_literal_constant_named = (
    ~sign
    + (
        significand + ~(exponent_letter + exponent)
        | digit_string + exponent_letter + exponent
    )
).named("value") + ~("_" + kind_param_named)
signed_real_literal_constant = ~sign + real_literal_constant

named_constant = name
real_part = signed_int_literal_constant | signed_real_literal_constant | named_constant
imag_part = real_part
complex_literal_constant = "(" + real_part + "," + imag_part + ")"

a_n_rep_char = Pattern("<alpha-numeric-rep-char>", r"\w")
rep_char = Pattern("<rep-char>", r".")
char_literal_constant = ~(kind_param + "_") + (
    "'" + ~~rep_char + "'" | '"' + ~~rep_char + '"'
)
a_n_char_literal_constant_named1 = ~(kind_param_named + "_") + (
    ~~~("'" + ~~a_n_rep_char + "'")
).named("value")
a_n_char_literal_constant_named2 = ~(kind_param_named + "_") + (
    ~~~('"' + ~~a_n_rep_char + '"')
).named("value")

logical_literal_constant = (r"[.]\s*(TRUE|FALSE)\s*[.]" + ~("_" + kind_param)).flags(
    re.I
)
logical_literal_constant_named = Pattern(
    "<value>", r"[.]\s*(TRUE|FALSE)\s*[.]", flags=re.I
).named() + ~("_" + kind_param_named)
literal_constant = (
    int_literal_constant
    | real_literal_constant
    | complex_literal_constant
    | logical_literal_constant
    | char_literal_constant
    | boz_literal_constant
)
constant = literal_constant | named_constant
int_constant = int_literal_constant | boz_literal_constant | named_constant
char_constant = char_literal_constant | named_constant

# assume that replace_string_map is applied:
part_ref = name + ~((r"[(]" + name + r"[)]"))
data_ref = part_ref + ~~~(r"[%]" + part_ref)
primary = constant | name | data_ref | (r"[(]" + name + r"[)]")

power_op = Pattern("<power-op>", r"(?<![*])[*]{2}(?![*])")
mult_op = Pattern("<mult-op>", r"(?<![*])[*](?![*])|(?<![/])[/](?![/])")
add_op = Pattern("<add-op>", r"[+-]")
concat_op = Pattern("<concat-op>", r"(?<![/])[/]\s*[/](?![/])")
rel_op = Pattern(
    "<rel-op>",
    r"[.]\s*EQ\s*[.]|[.]\s*NE\s*[.]|[.]\s*LT\s*[.]|[.]\s*LE\s*[.]|"
    r"[.]\s*GT\s*[.]|[.]\s*GE\s*[.]|[=]{2}|/[=]|[<][=]|[<]|[>][=]|[>]",
    flags=re.I,
)
not_op = Pattern("<not-op>", r"[.]\s*NOT\s*[.]", flags=re.I)
and_op = Pattern("<and-op>", r"[.]\s*AND\s*[.]", flags=re.I)
or_op = Pattern("<or-op>", r"[.]\s*OR\s*[.]", flags=re.I)
equiv_op = Pattern("<equiv-op>", r"[.]\s*EQV\s*[.]|[.]\s*NEQV\s*[.]", flags=re.I)
percent_op = Pattern("<percent-op>", r"%", flags=re.I)
intrinsic_operator = (
    power_op
    | mult_op
    | add_op
    | concat_op
    | rel_op
    | not_op
    | and_op
    | or_op
    | equiv_op
)
extended_intrinsic_operator = intrinsic_operator

defined_unary_op = Pattern("<defined-unary-op>", r"[.]\s*[A-Z]+\s*[.]", flags=re.I)
defined_binary_op = Pattern("<defined-binary-op>", r"[.]\s*[A-Z]+\s*[.]", flags=re.I)
defined_operator = defined_unary_op | defined_binary_op | extended_intrinsic_operator
abs_defined_operator = abs(defined_operator)
defined_op = Pattern("<defined-op>", "[.][A-Z]+[.]", flags=re.I)
abs_defined_op = abs(defined_op)

non_defined_binary_op = intrinsic_operator | logical_literal_constant

label = Pattern("<label>", r"\d{1,5}")
abs_label = abs(label)

keyword = name
keyword_equal = keyword + "="

abs_constant = abs(constant)
abs_literal_constant = abs(literal_constant)
abs_int_literal_constant = abs(int_literal_constant)
abs_signed_int_literal_constant = abs(signed_int_literal_constant)
abs_signed_int_literal_constant_named = abs(signed_int_literal_constant_named)
abs_int_literal_constant_named = abs(int_literal_constant_named)
abs_real_literal_constant = abs(real_literal_constant)
abs_signed_real_literal_constant = abs(signed_real_literal_constant)
abs_signed_real_literal_constant_named = abs(signed_real_literal_constant_named)
abs_real_literal_constant_named = abs(real_literal_constant_named)
abs_complex_literal_constant = abs(complex_literal_constant)
abs_logical_literal_constant = abs(logical_literal_constant)
abs_char_literal_constant = abs(char_literal_constant)
abs_boz_literal_constant = abs(boz_literal_constant)
abs_name = abs(name)
abs_a_n_char_literal_constant_named1 = abs(a_n_char_literal_constant_named1)
abs_a_n_char_literal_constant_named2 = abs(a_n_char_literal_constant_named2)
abs_logical_literal_constant_named = abs(logical_literal_constant_named)
abs_binary_constant = abs(binary_constant)
abs_octal_constant = abs(octal_constant)
abs_hex_constant = abs(hex_constant)

intrinsic_type_name = Pattern(
    "<intrinsic-type-name>",
    r"(INTEGER|REAL|COMPLEX|LOGICAL|CHARACTER|DOUBLE\s*COMPLEX|"
    r"DOUBLE\s*PRECISION|BYTE)",
    flags=re.I,
)
abs_intrinsic_type_name = abs(intrinsic_type_name)
double_complex_name = Pattern(
    "<double-complex-name>", r"DOUBLE\s*COMPLEX", flags=re.I, value="DOUBLE COMPLEX"
)
double_precision_name = Pattern(
    "<double-precision-name>",
    r"DOUBLE\s*PRECISION",
    flags=re.I,
    value="DOUBLE PRECISION",
)
abs_double_complex_name = abs(double_complex_name)
abs_double_precision_name = abs(double_precision_name)

access_spec = Pattern("<access-spec>", r"PUBLIC|PRIVATE", flags=re.I)
abs_access_spec = abs(access_spec)

implicit_none = Pattern(
    "<implicit-none>", r"IMPLICIT\s*NONE", flags=re.I, value="IMPLICIT NONE"
)
abs_implicit_none = abs(implicit_none)

attr_spec = Pattern(
    "<attr-spec>",
    r"(ALLOCATABLE|ASYNCHRONOUS|EXTERNAL|INTENT|INTRINSIC|"
    "OPTIONAL|PARAMETER|POINTER|PROTECTED|SAVE|TARGET|VALUE|VOLATILE)",
    flags=re.I,
)
abs_attr_spec = abs(attr_spec)

attr_spec_f08 = Pattern(
    "<attr-spec>",
    r"({})".format(
        "|".join(  # extend attr_spec with attribute CONTIGUOUS
            sorted(attr_spec.pattern.strip("()").split("|") + ["CONTIGUOUS"])
        )
    ),
    flags=re.I,
)
abs_attr_spec_f08 = abs(attr_spec_f08)

dimension = Pattern("<dimension>", r"DIMENSION", flags=re.I)
abs_dimension = abs(dimension)

intent = Pattern("<intent>", r"INTENT", flags=re.I)
abs_intent = abs(intent)

intent_spec = Pattern("<intent-spec>", r"INOUT|IN|OUT", flags=re.I)
abs_intent_spec = abs(intent_spec)

function = Pattern("<function>", r"FUNCTION", flags=re.I)
subroutine = Pattern("<subroutine>", r"SUBROUTINE", flags=re.I)

select_case = Pattern(
    "<select-case>", r"SELECT\s*CASE", flags=re.I, value="SELECT CASE"
)
abs_select_case = abs(select_case)



# Copyright (c) 2020-2021 Science and Technology Facilities Council.
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R466: this file tests the support for the
various forms of array constructor specifier.

Note that the tostr method is not yet tested.

"""

import pytest
from fparser.two.utils import NoMatchError
from fparser.two import Fortran2003


@pytest.mark.usefixtures("f2003_create")
def test_zero_size_array_constructor():
    """Test that we can parse a valid, zero-size array constructor."""
    fcode = "integer ::"
    ast = Fortran2003.Ac_Spec(fcode)
    assert isinstance(ast, Fortran2003.Ac_Spec)
    assert isinstance(ast.children[0], Fortran2003.Intrinsic_Type_Spec)


def test_int_literals_array_constructor():
    """Test when a simple list of integer literals is provided as the
    content of the constructor."""
    fcode = "1, 2, 3"
    ast = Fortran2003.Ac_Spec(fcode)
    assert isinstance(ast, Fortran2003.Ac_Value_List)


@pytest.mark.usefixtures("fake_symbol_table")
def test_expr_list_array_constructor():
    """Test when the provided content consists of expressions."""
    fcode = "ACOS(-1.0), SIN(1.0), 1.0+3.0"
    ast = Fortran2003.Ac_Spec(fcode)
    assert isinstance(ast, Fortran2003.Ac_Value_List)


@pytest.mark.usefixtures("f2003_create")
def test_array_spec_char_len():
    """Test with a specifier that specifies a length type parameter."""
    fcode = "CHARACTER(LEN=7) :: 'Takata', 'Tanaka', 'Hayashi'"
    ast = Fortran2003.Ac_Spec(fcode)
    assert isinstance(ast, Fortran2003.Ac_Spec)
    assert isinstance(ast.children[0], Fortran2003.Intrinsic_Type_Spec)
    assert "CHARACTER(LEN = 7) :: 'Takata', 'Tanaka', 'Hayashi'" in str(ast)


@pytest.mark.usefixtures("f2003_create")
def test_array_spec_no_match():
    """Check that incorrect content is not matched."""
    fcode = "call hello()"
    with pytest.raises(NoMatchError):
        Fortran2003.Ac_Spec(fcode)



# Modified work Copyright (c) 2018-2023 Science and Technology
# Facilities Council.
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""This file provides utilities to create a Fortran parser suitable
for a particular standard."""

import inspect
import logging
import sys
from fparser.two.symbol_table import SYMBOL_TABLES


def get_module_classes(input_module):
    """
    Return all classes local to a module.

    :param module input_module: the module containing the classes.

    :returns: list of class names and types.
    :rtype: List[Tuple[str, type]]

    """
    module_cls_members = []
    module_name = input_module.__name__
    # First find all classes in the module. This includes imported classes.
    all_cls_members = inspect.getmembers(sys.modules[module_name], inspect.isclass)
    # next only keep classes that are specified in the module.
    for name, cls in all_cls_members:
        if cls.__module__ == module_name:
            module_cls_members.append((name, cls))
    return module_cls_members


class ParserFactory:
    """Creates a parser suitable for the specified Fortran standard."""

    def create(self, std=None):
        """Creates a class hierarchy suitable for the specified Fortran
        standard. Also sets-up the list of classes that define scoping
        regions in the global SymbolTables object and clears any existing
        symbol table information.

        :param str std: the Fortran standard. Choices are 'f2003' or \
                        'f2008'. 'f2003' is the default.
        :return: a Program class (not object) for use with the Fortran reader
        :rtype: :py:class:`fparser.two.Fortran2003.Program`

        :raises ValueError: if the supplied value for the std parameter \
                            is invalid

        For example:

        >>> from fparser.two.parser import ParserFactory
        >>> f2003_parser = ParserFactory().create()
        >>> f2003_parser = ParserFactory().create(std='f2003')
        >>> f2008_parser = ParserFactory().create(std='f2008')
        >>> # Assuming that a reader has already been created ...
        >>> ast = f2008_parser(reader)
        >>> print(ast)

        """
        # Clear any existing symbol tables.
        SYMBOL_TABLES.clear()

        # find all relevant classes in our Fortran2003 file as we
        # always need these.
        # pylint: disable=import-outside-toplevel
        from fparser.two import Fortran2003

        f2003_cls_members = get_module_classes(Fortran2003)
        if not std:
            # default to f2003.
            std = "f2003"

        if std == "f2003":
            # we already have our required list of classes so call _setup
            # to setup our class hierarchy.
            self._setup(f2003_cls_members)
            # The class hierarchy has been set up so return the top
            # level class that we start from when parsing Fortran code.
            return Fortran2003.Program
        if std == "f2008":
            # we need to find all relevent classes in our Fortran2003
            # and Fortran2008 files and then ensure that where classes
            # have the same name we return the Fortran2008 class
            # i.e. where Fortran2008 extends Fortran2003 we return
            # Fortran2008.
            # First find all Fortran2008 classes.
            from fparser.two import Fortran2008

            f2008_cls_members = inspect.getmembers(
                sys.modules[Fortran2008.__name__], inspect.isclass
            )

            # next add in Fortran2003 classes if they do not already
            # exist as a Fortran2008 class.
            f2008_class_names = [i[0] for i in f2008_cls_members]
            for local_cls in f2003_cls_members:
                if local_cls[0] not in f2008_class_names:
                    f2008_cls_members.append(local_cls)
            # we now have our required list of classes so call _setup
            # to setup our class hierarchy.
            self._setup(f2008_cls_members)
            # The class hierarchy has been set up so return the top
            # level class that we start from when parsing Fortran
            # code. Fortran2008 does not extend the top level class so
            # we return the Fortran2003 one.
            return Fortran2003.Program

        raise ValueError(f"'{std}' is an invalid standard")

    def _setup(self, input_classes):
        """Perform some Python magic to create the connections between classes
        and populate the baseclass with this information. This has
        been lifted from the original implementation and no attempt
        has been made to tidy up the code, other than making it
        conformant to the coding rules.

        :param list input_classes: a list of tuples each containing a \
        class name and a class.

        """
        # pylint: disable=import-outside-toplevel
        from fparser.two import Fortran2003

        class_type = type(Fortran2003.Base)

        # Reset subclasses dictionary in case this function has been
        # called before. If this is not done then multiple calls to
        # the ParserFactory create method may not work correctly.
        Fortran2003.Base.subclasses = {}
        base_classes = {}

        for _, cls in input_classes:
            # ?? classtype is set to Base so why have issubclass?
            if (
                isinstance(cls, class_type)
                and issubclass(cls, Fortran2003.Base)
                and not cls.__name__.endswith("Base")
            ):
                base_classes[cls.__name__] = cls

        # OPTIMIZE subclass_names tree.
        #
        def _closest_descendants_with_match(clsname):
            """
            Starting at the named class, searches down the tree defined by the
            classes named in the `subclass_names` list to find the closest that
            have `match` methods. If the current class does not have a
            `match` method then this method is called again for each of
            the classes in its `subclass_names` list.

            :param str clsname: The name of the class from which to search.

            :returns: names of 'nearest' subclasses with `match` methods.
            :rtype: List[str | NoneType]

            """
            if clsname not in base_classes:
                error_string = f"Not implemented: {clsname}"
                logging.getLogger(__name__).debug(error_string)
                return []
            # remove this code when all classes are implemented.
            cls = base_classes[clsname]
            if hasattr(cls, "match"):
                # This class has a `match` method so no need to search further
                # down the tree.
                return [clsname]
            # clsname doesn't have a `match` method so we look at each of its
            # subclasses and find the nearest class in each that does have a
            # `match` method.
            bits = []
            for names in getattr(cls, "subclass_names", []):
                list1 = _closest_descendants_with_match(names)
                for names1 in list1:
                    if names1 not in bits:
                        bits.append(names1)
            return bits

        # Dict in which to store optimised list of subclass names for each cls.
        local_subclass_names = {}

        for cls in base_classes.values():
            if not hasattr(cls, "subclass_names"):
                continue
            # The optimised list of subclass names will only include subclasses
            # that have `match` methods.
            opt_subclass_names = []
            for names in cls.subclass_names:
                for names1 in _closest_descendants_with_match(names):
                    if names1 not in opt_subclass_names:
                        opt_subclass_names.append(names1)
            local_subclass_names[cls] = opt_subclass_names[:]

        # Now that we've optimised the list of subclass names for each class,
        # use this information to initialise the Base.subclasses dictionary:
        for clsname, cls in base_classes.items():
            if not hasattr(cls, "subclass_names"):
                message = f"{clsname} class is missing subclass_names list"
                logging.getLogger(__name__).debug(message)
                continue
            subclass_names = local_subclass_names.get(cls, [])
            try:
                bits = Fortran2003.Base.subclasses[clsname]
            except KeyError:
                Fortran2003.Base.subclasses[clsname] = bits = []
            for name in subclass_names:
                if name in base_classes:
                    bits.append(base_classes[name])
                else:
                    message = f"{name} not implemented needed by {clsname}"
                    logging.getLogger(__name__).debug(message)

        # Double-check that all required classes have been constructed.
        for cls in base_classes.values():
            subclass_names = local_subclass_names.get(cls, [])
            use_names = getattr(cls, "use_names", [])
            for name in use_names + subclass_names:
                if name not in base_classes:
                    message = f"{name} not defined, used by {cls.__name__}"
                    logging.getLogger(__name__).debug(message)



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""Test Fortran 2008 rule R824

    action-term-do-construct is label-do-stmt
                                      do-body
                                      do-term-action-stmt

The only difference to F2003 rule R835 is that we force this rule to
use the F2008 version of label-do-stmt

"""

import pytest

from fparser.api import get_reader
from fparser.two.Fortran2008 import Action_Term_Do_Construct


@pytest.mark.usefixtures("f2008_create")
def test_concurrent():
    """Test that the Fortran2008 version supports do concurrent."""
    code = "DO 10 CONCURRENT (i = 1 : 20)\n  a(i) = 0.0\n10 b(i) = 1.0"
    reader = get_reader(code)
    obj = Action_Term_Do_Construct(reader)
    assert isinstance(obj, Action_Term_Do_Construct)
    assert str(obj) == code


def test_functional(f2008_parser):
    """The 2008 version of the Action_Term_Do_Construct class is only
    added to make sure that that a labelled do concurrent (where the
    label is not attached to a continue) is parsed in f2008. Therefore
    add a functional test to make sure this class does its job.

    """
    code = (
        "PROGRAM test\n"
        "  INTEGER :: i\n"
        "  REAL :: a(20), b(20)\n"
        "  DO 10 CONCURRENT (i = 1 : 20)\n"
        "    a(i) = 0.0\n"
        "10 b(i) = 1.0\n"
        "END PROGRAM"
    )
    tree = f2008_parser(get_reader(code))
    assert str(tree) == code



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing tests for Fortran2003 where_construct_stmt rule R745.

R745 where-construct-stmt is [where-construct-name:] WHERE ( mask-expr )

"""

import pytest

from fparser.common.sourceinfo import FortranFormat
from fparser.one.parsefortran import FortranParser
from fparser.common.readfortran import FortranStringReader


@pytest.mark.parametrize("name", ["", "name"])
def test_where_construct_stmt(name):
    """Test that the Fortran2003 where_construct_stmt rule R745 works
    correctly. Test with and without the optional
    where-construct-name.

    R745 where-construct-stmt is [where-construct-name:] WHERE ( mask-expr )

    """
    space = ""
    colon = ""
    if name:
        colon = ":"
        space = " "
    code = (
        f"  SUBROUTINE columnwise_code()\n"
        f"    {name}{colon}{space}WHERE ( col_mat(:,:,cell) /= 0.0_r_solve )\n"
        f"      col_mat(:,:,cell) = 1.0_r_solver/col_mat(:,:,cell)\n"
        f"    END WHERE{space}{name}\n"
        f"  END SUBROUTINE columnwise_code"
    )
    reader = FortranStringReader(code)
    reader.set_format(FortranFormat(True, False))
    parser = FortranParser(reader)
    parser.parse()
    mod = parser.block.content[0]
    assert str(mod) == code



#!/usr/bin/env python
# Modified work Copyright (c) 2017-2019 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""Python script with command line options which calls the Fortran
File Reader with the supplied filename(s) and outputs the reader's
representation of the code(s).

"""

import sys
import logging
from fparser.scripts.script_options import set_read_options

logging.basicConfig()

try:
    from iocbio.optparse_gui import OptionParser
except ImportError:
    from optparse import OptionParser


def runner(_, options, args):
    """Call the Fortran File reader for each filename in args and print
    out its content.

    :param options: command line argument information from the options \
    parser
    :type options: :py:class:`optparse.Values`
    :param args: a list of Fortran filepaths
    :type args: list of str

    :raises NotImplementedError: if the task option is not set to \
    "show".

    """
    from fparser.common.readfortran import FortranFileReader

    for filename in args:
        reader = FortranFileReader(filename)
        if options.task == "show":
            for item in reader:
                print(item)
                sys.stdout.flush()
        else:
            raise NotImplementedError(
                "The task option '{0}' is invalid. Currently only "
                "'show' is supported.".format(repr(options.task))
            )


def main():
    """Check input options then call the runner function."""
    parser = OptionParser()
    set_read_options(parser)
    options, args = parser.parse_args()
    runner(parser, options, args)


if __name__ == "__main__":
    main()



# Modified work Copyright (c) 2017-2022 Science and Technology
# Facilities Council.
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R504: entity-decl"""

from fparser.two.Fortran2003 import Entity_Decl, Name

import pytest


def test_entity_decl_repr():
    tcls = Entity_Decl
    obj = tcls("a(1)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "a(1)"
    assert (
        repr(obj) == "Entity_Decl(Name('a'), Explicit_Shape_Spec_List(',', "
        "(Explicit_Shape_Spec(None, Int_Literal_Constant('1', None)),)), "
        "None, None)"
    )


@pytest.mark.parametrize(
    ("declaration, expected_str"),
    [
        ("a(1)*(3)", "a(1)*(3)"),
        ("a(1)*(3) = 2", "a(1)*(3) = 2"),
        ("a = 2", "a = 2"),
        ("a=2", "a = 2"),
        ('a = "abc "', 'a = "abc "'),
        ("a = .true.", "a = .TRUE."),
    ],
)
def test_entity_decl_str(declaration, expected_str):
    """Test the string representations of various entity declarations"""
    obj = Entity_Decl(declaration)
    assert isinstance(obj, Entity_Decl), repr(obj)
    assert str(obj) == expected_str


def test_entity_decl_name():  # 504
    """Test we can get the name of an entity declaration"""
    obj = Entity_Decl("a(1) = 2")
    assert obj.get_name() == Name("a")



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Specification_Part constraint C1112
"""

from fparser.two.Fortran2003 import Specification_Part, Use_Stmt, Import_Stmt
from fparser.two.Fortran2008.declaration_construct_c1112 import (
    Declaration_Construct_C1112,
)
from fparser.two.Fortran2008.implicit_part_c1112 import Implicit_Part_C1112
from fparser.two.utils import BlockBase


class Specification_Part_C1112(Specification_Part):  # C1112
    """Fortran 2008 constraint C1112
    C1112 A submodule specification-part shall not contain a
    format-stmt, entry-stmt, or stmt-function-stmt.

    These statements are found in the following rule hierarchy

    format-stmt Specification_Part/implicit_part/implicit_part_stmt/format_stmt
                Specification_Part/declaration_construct/format_stmt
    entry-stmt Specification_Part/implicit_part/implicit_part_stmt/entry_stmt
               Specification_Part/declaration_construct/entry_stmt
    stmt-function-stmt Specification_Part/declaration_construct/
                       stmt-function-stmt

    Therefore we need to specialise implicit_part, implicit_part_stmt
    and declaration_construct

    """

    use_names = [
        "Use_Stmt",
        "Import_Stmt",
        "Implicit_Part_C1112",
        "Declaration_Construct_C1112",
    ]

    @staticmethod
    def match(reader):
        """Check whether the input matches the rule

        param reader: the fortran file reader containing the line(s)
                      of code that we are trying to match
        :type reader: :py:class:`fparser.common.readfortran.FortranFileReader` \
            | :py:class:`fparser.common.readfortran.FortranStringReader`

        :returns: `tuple` containing a single `list` which contains
                 instance of the classes that have matched if there is
                 a match or `None` if there is no match
        :rtype: Optional[Tuple[List[:py:class:`fparser.two.utils.Base`]]]
        """
        return BlockBase.match(
            None,
            [Use_Stmt, Import_Stmt, Implicit_Part_C1112, Declaration_Construct_C1112],
            None,
            reader,
        )
